# Copyright Haven Selph 2022, MIT License, see LICENSE.md file for more info.
import inspect
from .parser import *


class Command:
    def __init__(self, funct: callable, names: list[str, ...] or tuple[str, ...], does: str=None, help_msg: str=None) -> None:
        """
        Generates a new command object to be used in Commands.

        :param funct: Function to be called when calling this object.
        :param names: All names for this object to be referred to as.
        :param does: A short description of what this command does.
        :param help_msg: A long description of what this command does.
        """
        self.__funct = funct
        self.__names = names
        self.does = does
        self.help_msg = help_msg

    @property
    def funct(self) -> callable:
        """Returns the function of this command."""
        return self.__funct

    @property
    def parameters(self) -> list[str, ...] or tuple[str, ...]:
        """Returns all parameters of this commands function as a tuple."""
        return [arg[0] for arg in inspect.signature(self.funct).parameters]

    @property
    def names(self) -> list[str, ...] or tuple[str, ...]:
        """Returns all names of this command."""
        return self.__names

    @property
    def name(self) -> str:
        """Returns names[0] of this command."""
        return self.__names[0]

    def usage(self, name: str=None, aliases_sep: str=" | ", aliases_wrap: list[str, str] or tuple[str, str]=("<",">"), param_sep: str=" | ", param_wrap: list[str, str] or tuple[str, str]=("<",">"), setup: str="|{name}: {does}\n|{parameters}") -> str:
        """
        Generate a string to show command usage dynamically. Use setup to customize how it looks.
        Example Setup String: "{name}: {does}\n{parameters}"

        {name} = name parameter

        {aliases} = self.names (does not include name, and formatted using provided parameters)

        {parameters} = self.parameters (formatted using provided parameters)

        {does} = self.does

        {help_msg} = self.help_msg


        :param name: Which name to use as the main name of the command. Defaults to commands main name.
        :param aliases_sep: Text to place between each alias.
        :param aliases_wrap: Tuple where [0] is placed on the left and [1] on the right of each alias.
        :param param_sep: Text to place between each parameter.
        :param param_wrap: Tuple where [0] is placed on the left and [1] on the right of each alias.
        :param setup: Custom setup for string.


        :return: Returns a generated string based on parameters passed.
        """
        name = name if name in self.names else self.name
        parameters = param_sep.join([param_wrap[0] + param + param_wrap[1] for param in self.parameters])
        aliases = aliases_sep.join([aliases_wrap[0] + str(_name) + aliases_wrap[1] for _name in self.names if _name!=name])
        return setup.format(
            name=name,
            aliases=aliases or "",
            does=self.does or "",
            parameters=parameters,
            help_msg=self.help_msg or ""
        )

    def __call__(self, *args, **kwargs):
        return self.funct(*args, **kwargs)


class Commands(dict):
    def __init__(self, /, register_help_command: bool=True, register_exit_command: bool=True, register_aliases_command: bool=True) -> None:
        """Create a commands dictionary to register commands to."""
        self.names_no_aliases = []

        if register_help_command:
            self.__help_command = self.add_command("help", "info", does="Displays information on a provided command, or all registered commands.")(self.__help_command)

        if register_exit_command:
            self.add_command("exit", "quit", does="Exits the script using Python's built-in 'exit()' method.")(exit)

        if register_aliases_command:
            self.__aliases = self.add_command("aliases", "names", does="Displays all names pointing to a provided command.")(self.__aliases)
        super().__init__()

    def __help_command(self, command: str=None) -> None:
        if command:
            setup = "Command: {name} - {does}" if self.get(command).does else "Command: {name}"
            setup += "\nParameters: {parameters}" if len(self.get(command).parameters)>0 else "\nParameters: [no parameters]"
            setup += "\n{help_msg}" if self.get(command).parameters else ""
            print(self.get(command).usage(command, param_sep=" ", param_wrap=("[", "]"), setup=setup), end="\n" * 2)
        else:
            for command in self.names_no_aliases:
                setup = "Command: {name} - {does}" if self.get(command).does else "Command: {name}"
                setup += "\nParameters: {parameters}" if len(self.get(command).parameters)>0 else "\nParameters: [no parameters]"
                setup += "\n{help_msg}" if self.get(command).parameters else ""
                print(self.get(command).usage(command, param_sep=" ", param_wrap=("[","]"), setup=setup), end="\n"*2)

    def __aliases(self, command: str):
        setup = "Command: {name}"
        setup += "Aliases: {aliases}" if len(self.get(command).names)>1 else "Aliases: Command has no aliases."
        print(self.get(command).usage(aliases_sep=", ", aliases_wrap=("",""), setup="Command: {name}\nAliases: {aliases}"), end="\n"*2)

    def __add_command(self, command: Command) -> Command:
        for name in command.names:
            if name in self:
                raise self.CommandExists(name)
            else:
                self[name] = command
        self.names_no_aliases.append(command.name)
        return command

    def add_command(self, *names: str, does: str=None, help_msg: str=None) -> callable:
        def wrapper(funct: callable) -> Command:
            return self.__add_command(Command(funct, names, does=does, help_msg=help_msg))

        return wrapper

    def execute(self, text: str) -> any:
        _out = self.parse_string(text)
        return self.execute_no_parse(*_out[0], **_out[1])

    def execute_no_parse(self, *args, **kwargs) -> any:
        if not args:
            raise self.CommandNotFound("No command provided.")
        else:
            if args[0] in self:
                return self.get(args[0])(*args[1:], **kwargs)
            else:
                raise self.CommandNotFound(f"Command '{args[0]}' not found.")

    @classmethod
    def parse_string(cls, string: str="") -> tuple[list, dict]:
        try:
            _out = parse(string)
        except Tokenizer.UnexpectedToken as er:
            raise cls.ParseError(repr(er))
        except Tokenizer.UnexpectedEOF as er:
            raise cls.ParseError(repr(er))
        return _out

    def get(self, key) -> Command:  # Included so I can typehint .get()
        return self[key]

    class PyCommandsToolException(Exception):
        """A generic exception to be inherited by all custom exceptions in PyCommandsTool."""
        def __init__(self, *values: str) -> None:
            super().__init__(*values)

    class CommandExists(PyCommandsToolException):
        """An exception thrown when a command name provided is already registered to the Commands object."""

    class CommandNotFound(PyCommandsToolException):
        """An exception thrown when a command name does not exist."""

    class ParseError(PyCommandsToolException):
        """An exception thrown when parse cannot properly interpret a string passed to it."""


if __name__ == "__main__":
    cmd = Commands()

    @cmd.add_command("add", "sum", does="Adds two numbers together.")
    def add(a, b):
        print(a + b)

    while True:
        try:
            cmd.execute(input(">>> "))
        except cmd.PyCommandsToolException as e:
            print(e)



import re


def parse(text: str, protected: bool=False, debug=False) -> tuple[list, dict]:
    keywords = {}

    def helper(_token: dict) -> any:
        if debug:
            print(f"Found a {_token['type']}")
        match _token:
            case {"type": "word"}:
                return _token["value"]
            case {"type": "string"}:
                return _token["value"]
            case {"type": "float"}:
                return float(_token["value"])
            case {"type": "integer"}:
                return int(_token["value"])
            case {"type": "list"}:
                return [helper(_tkn) for _tkn in _token["value"]]
            case {"type": "tuple"}:
                return tuple([helper(_tkn) for _tkn in _token["value"]])
            case {"type": "dict"}:
                return {_tkn["left"]: helper(_tkn["right"]) for _tkn in _token["value"]}
            case {"type": "assignment"}:
                keywords[_token["left"]] = helper(_token["right"])

    if debug:
        print(f"Parsing string: {text}")

    _Tokenizer = Tokenizer(text)
    _out = []
    try:
        if debug:
            print("Getting all tokens")
        token = _Tokenizer.getNextToken()

        while token["type"] != "EOF":
            if debug:
                print(f"Got Token: {token}")
            _out.append(helper(token))
            token = _Tokenizer.getNextToken()
        if debug:
            print(f"Returning: {_out}", end="\n\n")
        return _out, keywords
    except Tokenizer.UnexpectedToken as e:
        if debug:
            print(e)
        if protected:
            return [],{}
        else:
            raise e
    except Tokenizer.UnexpectedEOF as e:
        if debug:
            print(e)
        if protected:
            return [],{}
        else:
            raise e


class Tokenizer:
    class UnexpectedEOF(Exception):
        pass

    class UnexpectedToken(Exception):
        pass

    def __init__(self, text: str) -> None:
        """Initializes tokenizer with text"""
        self.text = text
        self._cursor = 0

    @property
    def notEOF(self) -> bool:
        """Returns True if not at end of file"""
        return self._cursor<len(self.text)

    @property
    def cursor(self) -> int:
        """Returns current cursor position"""
        return self._cursor

    @property
    def atCursor(self) -> str:
        """Returns current character at cursor"""
        if self.notEOF:
            return self.text[self._cursor]
        else:
            return ""

    @property
    def fromCursor(self) -> str:
        """Returns text from cursor"""
        return self.text[self._cursor:]

    def matchCursor(self, exp: str) -> bool:
        """Returns True if cursor matches exp"""
        return re.match(exp, self.atCursor) is not None

    def mvCursor(self, i: int=1) -> int:
        """Moves cursor by i and returns new cursor position"""
        self._cursor += i
        return self.cursor

    def consume(self, exp=r"\s") -> str:
        """Returns next character that doesn't match exp"""
        while re.match(exp, self.atCursor):
            self.mvCursor()
        return self.atCursor

    def getNextToken(self) -> dict:  # Recursive Descent Parser
        """Returns next token"""

        while self.notEOF:  # While not at end of file
            self.consume()
            if not self.notEOF:
                return {"type": "EOF", "value": None}
            if self.matchCursor(r"[a-zA-Z]"):  # WORD
                _out = ""
                while self.notEOF and not self.matchCursor(r"\s"):
                    if self.matchCursor(r"[a-zA-Z]|[0-9]"):  # Checks if character is a letter or number
                        _out += self.atCursor
                        self.mvCursor()
                    elif self.matchCursor(r"="):  # Checks if character is an assignment
                        self.mvCursor()
                        _tmp = self.getNextToken()
                        if _tmp["type"] in ("float", "integer", "string", "list", "tuple", "dict"):  # If token is of valid type
                            return {"type": "assignment", "left": _out, "right": _tmp}
                        elif _tmp["type"] == "EOF":  # If token is EOF raise error
                            raise self.UnexpectedEOF("Unexpected EOF")
                        else:  # If token is invalid raise error
                            raise self.UnexpectedToken(f"Unexpected token '{self.atCursor}'")
                    elif self.matchCursor(r":"):  # Checks if character is a dictionary
                        self.mvCursor()
                        _tmp = self.getNextToken()
                        if _tmp["type"] in ("float", "integer", "string", "list", "tuple", "dict"):  # If token is of valid type
                            return {"type": "dict-assignment", "left": _out, "right": _tmp}
                        else:  # If token is invalid raise error
                            raise self.UnexpectedToken(f"Unexpected token '{self.atCursor}'")
                    else:  # If character is not valid raise error
                        raise self.UnexpectedToken(f"Unexpected character '{self.atCursor}'")
                self.mvCursor()
                return {"type": "word", "value": _out}
            elif self.matchCursor(r"[0-9]"):  # NUMBER
                _out = ""
                while self.notEOF and self.matchCursor(r"[0-9]|\."):
                    if self.matchCursor(r"[0-9]"):  # Checks if character is a number
                        _out += self.atCursor
                        self.mvCursor()
                    elif self.matchCursor(r"\."):  # Checks if character is a decimal
                        if "." in _out:  # Checks if there is already a decimal
                            raise self.UnexpectedToken(f"Invalid number '{self.atCursor}'")
                        _out += self.atCursor
                        self.mvCursor()
                    else:  # If character is not valid raise error
                        raise self.UnexpectedToken(f"Unexpected character '{self.atCursor}'")
                return {"type": "float" if "." in _out else "integer", "value": float(_out) if "." in _out else int(_out)}
            elif self.matchCursor(r"\"|\'"):  # STRING
                _out = ""
                _quote = self.atCursor
                self.mvCursor()
                while self.atCursor != _quote:  # Until character is a closing quote
                    if not self.notEOF:  # If at end of file raise error
                        raise self.UnexpectedEOF("Unexpected EOF")
                    if self.matchCursor(r"\\"):  # Checks if character is an escape character
                        self.mvCursor()
                        if self.matchCursor('n'):  # Checks if character is a newline
                            raise self.UnexpectedToken("Unexpected character '\\n'")
                    _out += self.atCursor
                    self.mvCursor()
                self.mvCursor()
                return {"type": "string", "value": _out}

            elif self.matchCursor(r"\("):  # TUPLE
                self.mvCursor()
                self.consume(" ")
                _out = [self.getNextToken()]
                while not self.matchCursor(r"\)"):  # Until character is a closing parenthesis
                    self.consume(" ")
                    if not self.notEOF:  # If at end of file raise error
                        raise self.UnexpectedToken("Unexpected EOF")
                    if self.matchCursor(r","):
                        self.mvCursor()
                        _tmp = self.getNextToken()
                        if _tmp["type"] in ("float", "integer", "string", "list", "tuple", "dict"):  # If token is of valid type
                            _out.append(_tmp)
                        else:  # If token is invalid raise error
                            raise self.UnexpectedToken(f"Unexpected token '{_tmp['type']}'")
                    else:
                        raise self.UnexpectedToken(f"Unexpected character '{self.atCursor}'")
                self.mvCursor()
                return {"type": "tuple", "value": _out}

            elif self.matchCursor(r"\["):  # LIST
                self.mvCursor()
                self.consume(" ")
                _out = [self.getNextToken()]
                while not self.matchCursor(r"\]"):  # Until character is a closing bracket
                    self.consume(" ")
                    if not self.notEOF:  # If at end of file raise error
                        raise self.UnexpectedEOF("Unexpected EOF")
                    if self.matchCursor(r","):
                        self.mvCursor()
                        _tmp = self.getNextToken()
                        if _tmp["type"] in ("float", "integer", "string", "list", "tuple", "dict"):  # If token is invalid raise error
                            _out.append(_tmp)
                        else:  # If token is invalid raise error
                            raise self.UnexpectedToken(f"Unexpected token '{_tmp['type']}'")
                    else:
                        raise self.UnexpectedToken(f"Unexpected character '{self.atCursor}'")
                self.mvCursor()
                return {"type": "list", "value": _out}

            elif self.matchCursor(r"\{"):  # DICTIONARY
                self.mvCursor()
                self.consume(" ")
                _out = [self.getNextToken()]
                while not self.matchCursor(r"\}"):  # Until character is a closing brace
                    self.consume(" ")
                    if not self.notEOF:  # If at end of file raise error
                        raise self.UnexpectedEOF("Unexpected EOF")
                    if self.matchCursor(r","):
                        self.mvCursor()
                        _tmp = self.getNextToken()
                        if _tmp["type"]=="dict-assignment":
                            _out.append(_tmp)
                        else:  # If token is not dict-assignment raise error
                            raise self.UnexpectedToken(f"Unexpected token '{_tmp['type']}'")
                    else:
                        raise self.UnexpectedToken(f"Unexpected character '{self.atCursor}'")
                self.mvCursor()
                return {"type": "dict", "value": _out}

            else:  # If character is not valid raise error
                raise self.UnexpectedToken(f"Unexpected character '{self.atCursor}'")
        return {"type": "EOF", "value": None}
